#ifndef NDS_SWIG_VERSION_HH
#define NDS_SWIG_VERSION_HH

#include <string>

namespace NDS_swig {

    inline std::string version()
    {
        return "0.16.7";
    }

}

#endif // NDS_SWIG_VERSION_HH
